#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <locale.h>
#include <ctype.h>

#include "ietraces.h"

#define ALIGNEMENT 0x10
#define READ_BLOCK_SIZE (ALIGNEMENT * 0x10000)
#define IE_BLOCK_SIZE 0x80
#define MAX_IE_BLOCKS 10

#if SIZEOF_OFF_T == 8
#ifdef _WIN32
#define PRIuOff "I64u"
#define PRIxOff "I64x"
#define PRIdOff "I64d"
#else
#define PRIuOff "llu"
#define PRIxOff "llx"
#define PRIdOff "lld"
#endif
#else
#define PRIuOff "u"
#define PRIxOff "x"
#define PRIdOff "d"
#endif

int
print_win_date(FILE *output, uint64_t value)
{
#define STARTYEAR 1601
#define SECTODAY (24L * 3600L)
#define NANOTOSEC 10000000
#define LEAPYEAR(year) (!((year) % 4) && (((year) % 100) || ! ((year) % 400)))
#define YEARSIZE(year) (LEAPYEAR(year)? 366 : 365)

     unsigned long long sec;
     unsigned long tm_nano;
     unsigned long dayclock;
     unsigned long dayno;
     int tm_sec, tm_min, tm_hour, tm_year, tm_mon, tm_mday;
     static unsigned int _wytab[2][12] =
          { {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
     {31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
     };

     sec = value / NANOTOSEC;
     tm_nano = value % NANOTOSEC;
     tm_year = STARTYEAR;
     dayclock = (unsigned long) (sec % SECTODAY);
     dayno = (unsigned long) (sec / SECTODAY);
     tm_sec = dayclock % 60;
     tm_min = (dayclock % 3600) / 60;
     tm_hour = dayclock / 3600;
     while (dayno >= YEARSIZE (tm_year)) {
          dayno -= YEARSIZE (tm_year);
          tm_year++;
     }
     tm_mon = 0;
     while (dayno >= _wytab[LEAPYEAR (tm_year)][tm_mon]) {
          dayno -= _wytab[LEAPYEAR (tm_year)][tm_mon];
          tm_mon++;
     }
     tm_mon++;
     tm_mday = dayno + 1;
/*     return fprintf (output, "%.2u/%.2u/%u %.2u:%.2u:%.2u.%.7lu",
              tm_mday, tm_mon, tm_year, tm_hour, tm_min, tm_sec, tm_nano); */
     return fprintf (output, "%.4u%.2u%.2u %.2u:%.2u:%.2u.%.7lu",
              tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec, tm_nano);

}

int
print_text(FILE *output, unsigned char *buffer, size_t maxsize)
{
     size_t curpos;
     int curchar, num;
     
     for (curpos = 0; curpos < maxsize && (curchar = (int)buffer[curpos]);
          curpos++) {
	  num++;
          if (!isprint(curchar) || curchar == 0xA || curchar == 0xD) {
	       fputc('.', output);
	  } else {
	       fputc(curchar, output);
	  }
     }
     return num;
}

void
parse_ie_record(unsigned char *buffer, size_t bsize, off_t offset)
{
#define PRN_TEXT_FIELD(buf, skip, max, off, tmp, ftext) \
     tmp = GET_U32((buf) + (skip)); \
     if ((tmp) < (max)) { \
          print_text(stdout, (buf) + (tmp), (max) - (tmp)); \
     } else { \
          fprintf(stderr, "Record %.4s at offset 0x%"PRIxOff" has "ftext" " \
	       "field outside of record\n", (buf), (off)); \
     } \
     fputc('\t', stdout);
     
     uint64_t iedate;
     size_t field_off;
     uint32_t ietype = GET_U32(buffer);

     /* IE5 record type offsets, for IE4 test bsize < 0x4C */
     if (ietype == URL_TYPE || ietype == LEAK_TYPE) {
          if (bsize < 0x48) {
               fprintf(stderr, "Record %.4s at offset 0x%"PRIxOff" is too "
	            "small (%u) for processing\n", buffer, offset, bsize);
	       return;
	  }
          fprintf(stdout, "%.4s\t", buffer);
          iedate = GET_U64(buffer + 8);
          print_win_date(stdout, iedate);
          fputc('\t', stdout);
          iedate = GET_U64(buffer + 0x10);
          print_win_date(stdout, iedate);
          fputc('\t', stdout);
	  /* offsets IE4 : 0x38, 0x40 et 0x48 respectivement */
	  PRN_TEXT_FIELD(buffer, 0x34, bsize, offset, field_off, "url");
	  PRN_TEXT_FIELD(buffer, 0x3C, bsize, offset, field_off, "filename");
	  PRN_TEXT_FIELD(buffer, 0x44, bsize, offset, field_off, "http_header");
     } else {
          if (bsize <= 0x10) {
               fprintf(stderr, "Record %.4s at offset 0x%"PRIxOff" is too "
	            "small (%u) for processing\n", buffer, offset, bsize);
	       return;
	  }
          fprintf(stdout, "%.4s\t\t\t", buffer);
	  print_text(stdout, buffer + 0x10, bsize - 0x10);
     }
     fputc('\n', stdout);
}

size_t
find_ie_record(unsigned char *buffer, off_t offset)
{
     size_t iesize = 0;
     uint64_t iedate = 0;
     uint32_t ietype = GET_U32(buffer);

     if (ietype != URL_TYPE && ietype != LEAK_TYPE && ietype != REDR_TYPE) {
          return 0;
     }
     if (ietype == URL_TYPE || ietype == LEAK_TYPE) {
          /* Elimine les dates aberrantes : on accepte soit 0, soit octet de
	   * poids fort à 1 */
	  if ((iedate = GET_U64(buffer + 8)) != 0 && buffer[0xF] != 1) {
	       fprintf(stderr, "Date ");
	       print_win_date(stderr, iedate);
	       fprintf(stderr, " (0x%016llx) for record type \"%.4s\" at offset"
	       " Ox%"PRIxOff" does not seem valid, skipping\n", iedate, buffer,
	       offset);
	       return 0;
	  }
     }

     iesize = GET_U32(buffer + 4);
     if (iesize > MAX_IE_BLOCKS) {
          fprintf(stderr, "IE blocks count (%u) seems too large,"
               " reducing to %u\n", iesize, MAX_IE_BLOCKS);
          iesize = MAX_IE_BLOCKS;
     }

     return iesize * IE_BLOCK_SIZE;
}

void
usage(FILE *file, char *name)
{
     fprintf(file, "Usage: %s filename\n", name);
}
int
main(int argc, char *argv[])
{
     int ret = EXIT_FAILURE;
     FILE *fd = NULL;
     unsigned char *block = NULL;
     off_t offset = 0;
     int not_eof = 1, count = 0;
     struct stat st;

     if (argc != 2) {
          usage(stderr, argv[0]);
	  goto err;
     }
     if ((block = malloc(READ_BLOCK_SIZE)) == NULL) {
          fprintf(stderr, "No memory available\n");
	  goto err;
     }
     if ((fd = fopen(argv[1], "r")) == NULL) {
          fprintf(stderr, "Can't open %s: %s\n", argv[1], strerror(errno));
	  goto err;
     }
     if (fstat(fileno(fd), &st) < 0) {
          fprintf(stderr, "Warning, can't stat file %s: %s\n",
	       argv[1], strerror(errno));
	  st.st_size = 0;
     }

     setlocale(LC_ALL, "");

     do {
	  size_t curpos = 0, delta = 0;
          size_t nread = fread((void *)block, 1, READ_BLOCK_SIZE, fd);
	  if (nread < READ_BLOCK_SIZE) {
                if (feof(fd)) {
		     not_eof = 0;
		} else {
		     fprintf(stderr, "Warning incomplete read at offset "
		          "0x%"PRIxOff, offset);
		     if (ferror(fd)) {
		          fprintf(stderr, ", aborting due to read error");
		          not_eof = 0;
		     }
		     fprintf(stderr, "\n");
		}
	  }

	  delta = nread % ALIGNEMENT;

	  for (curpos = 0; curpos < nread - delta ; curpos += ALIGNEMENT) {
	       size_t rec_size;
	       if((rec_size = find_ie_record(block+curpos, offset+curpos))) {
	            if ((rec_size + curpos) > nread) {
		         delta = max(delta, nread - curpos);
		    } else {
		         parse_ie_record(block+curpos, rec_size, offset+curpos);
		    }
	       }
	  }

	  offset += nread - delta;
	  if (count++ == 1000) {
	       count = 0;
	       if (st.st_size) {
	            fprintf(stderr, "%"PRIdOff"%% performed\n",
		         (offset * 100)/st.st_size);
	       }
	  }
	  if (delta && fseeko(fd, offset, SEEK_SET)) {
	       fprintf(stderr, "Can't set the file position at offset "
		    "0x%"PRIxOff": %s, arborting\n", offset, strerror(errno));
               not_eof = 0;
	  }
     } while (not_eof);

     ret = EXIT_SUCCESS;
err:
     if (fd)
          fclose(fd);
     if (block)
          free(block);
     exit(ret);
}
